/*
 * WMA compatible decoder
 * Copyright (c) 2002 The FFmpeg Project.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file wmadec.c
 * WMA compatible decoder.
 * This decoder handles Microsoft Windows Media Audio data, versions 1 & 2.
 * WMA v1 is identified by audio format 0x160 in Microsoft media files
 * (ASF/AVI/WAV). WMA v2 is identified by audio format 0x161.
 *
 * To use this decoder, a calling application must supply the extra data
 * bytes provided with the WMA data. These are the extra, codec-specific
 * bytes at the end of a WAVEFORMATEX data structure. Transmit these bytes
 * to the decoder using the extradata[_size] fields in AVCodecContext. There
 * should be 4 extra bytes for v1 data and 6 extra bytes for v2 data.
 */
 /*
  * Fixed point modifications by Derek Taubert (taubert@geeks.org) Feb 2005
  * Extra fixed point modifications by Peter McQuillan Jan 2006
  * Fixed point modifications complete by ell Feb 2007
 */

//#define DEBUG
#include <math.h>
#include "mdct.h"
#include "common.h"
#include "asfheader.h"

/* size of blocks */
#define BLOCK_MIN_BITS 7
#define BLOCK_MAX_BITS 11
#define BLOCK_MAX_SIZE (1 << BLOCK_MAX_BITS)

#define BLOCK_NB_SIZES (BLOCK_MAX_BITS - BLOCK_MIN_BITS + 1)

/* XXX: find exact max size */
#define HIGH_BAND_MAX_SIZE 16
#define NB_LSP_COEFS 10

/* XXX: is it a suitable value ? */
#define MAX_CODED_SUPERFRAME_SIZE 16384
#define MAX_CHANNELS 2
#define NOISE_TAB_SIZE 8192
#define LSP_POW_BITS 7
#define EXP_VLC_TAB_SIZE 63

//#define USEFLOAT
#undef USEFLOAT

/* Fixed point ops */
#if 1 //def NO_INTEGER_ROUNDING
#define IRINT(_v, _s)           ((_v) >> (_s))
#else
#define IRINT(_v, _s) ({                                                \
	typeof (_v) __v = (_v);                                             \
	typeof (__v) __r = ((typeof (__v))1) << ((_s) - 1);                 \
	__v += __r;                                                         \
	__v >> (_s);                                                        \
})
#endif

#if 1 // def NO_FPTOFX_ASSERT
#define FPTOFXS(_fp, _s)        ((int32_t)(lrintf((_fp) * (int32_t)((1LL<<(_s))-1))))
#define NFPTOFXS(_fp, _s)       ((int32_t)(lrintf((_fp) * (int32_t)(-(1LL<<(_s))))))
#define FPTOFXU(_fp, _s)        ((uint32_t)(llrintf((_fp) * (uint32_t)((1LL<<(_s))-1))))
#else
#include <assert.h>
#define FPTOFXS(_fp, _s) ({                                     \
	float __fp = (_fp);                                         \
	assert((__fp >= (float)-(1<<(31-(_s)))) &&                  \
		   (__fp <= (float)(1<<(31-(_s)))));                    \
	((int32_t)(lrintf(__fp * (int32_t)((1LL<<(_s))-1))));       \
})
#define NFPTOFXS(_fp, _s) ({                                    \
	float __fp = (_fp);                                         \
	assert((__fp >= (float)-(1<<(31-(_s)))) &&                  \
		   (__fp <= (float)(1<<(31-(_s)))));                    \
	((int32_t)(lrintf(__fp * (int32_t)(-(1LL<<(_s))))));        \
})
#define FPTOFXU(_fp, _s) ({                                     \
	float __fp = (_fp);                                         \
	assert((__fp >= 0.f) && (__fp <= (float)(1<<(32-(_s)))));   \
	((uint32_t)(llrintf(__fp * (uint32_t)((1LL<<(_s))-1))));    \
})
#endif

#define FXSMULT(_a, _b, _s)     IRINT((int64_t)(_a) * (_b), (_s))
#define FXUMULT(_a, _b, _s)     IRINT((uint64_t)(_a) * (_b), (_s)))

#if 0
#define CONDTRIG(_s) do {                                               \
	static int hit = 0;                                                 \
	if (!hit) {                                                         \
		hit = 1;                                                        \
		dprintf("\nLine %d, condition '%s' hit\n", __LINE__, (_s));     \
	}                                                                   \
} while (0)
#else
#define CONDTRIG(_s)
#endif

typedef struct WMADecodeContext {
	GetBitContext gb;
	int sample_rate;
	int nb_channels;
	int bit_rate;
	int version; /* 1 = 0x160 (WMAV1), 2 = 0x161 (WMAV2) */
	int block_align;
	int use_bit_reservoir;
	int use_variable_block_len;
	int use_exp_vlc;  /* exponent coding: 0 = lsp, 1 = vlc + delta */
	int use_noise_coding; /* true if perceptual noise is added */
	int byte_offset_bits;
	VLC exp_vlc;
	int exponent_sizes[BLOCK_NB_SIZES];
	uint16_t exponent_bands[BLOCK_NB_SIZES][25];
	int high_band_start[BLOCK_NB_SIZES]; /* index of first coef in high band */
	int coefs_start;               /* first coded coef */
	int coefs_end[BLOCK_NB_SIZES]; /* max number of coded coefficients */
	int exponent_high_sizes[BLOCK_NB_SIZES];
	int exponent_high_bands[BLOCK_NB_SIZES][HIGH_BAND_MAX_SIZE];
	VLC hgain_vlc;

	/* coded values in high bands */
	int high_band_coded[MAX_CHANNELS][HIGH_BAND_MAX_SIZE];
	int high_band_values[MAX_CHANNELS][HIGH_BAND_MAX_SIZE];

	/* there are two possible tables for spectral coefficients */
	VLC coef_vlc[2];
	uint16_t *run_table[2];
	uint16_t *level_table[2];
	/* frame info */
	int frame_len;       /* frame length in samples */
	int frame_len_bits;  /* frame_len = 1 << frame_len_bits */
	int nb_block_sizes;  /* number of block sizes */
	/* block info */
	int reset_block_lengths;
	int block_len_bits; /* log2 of current block length */
	int next_block_len_bits; /* log2 of next block length */
	int prev_block_len_bits; /* log2 of prev block length */
	int block_len; /* block length in samples */
	int block_num; /* block number in current frame */
	int block_pos; /* current position in frame */
	uint8_t ms_stereo; /* true if mid/side stereo mode */
	uint8_t channel_coded[MAX_CHANNELS]; /* true if channel is coded */
#ifdef USEFLOAT
	float exponents[MAX_CHANNELS][BLOCK_MAX_SIZE] __attribute__((aligned(16)));
	float max_exponent[MAX_CHANNELS];
#else
	uint32_t exponents[MAX_CHANNELS][BLOCK_MAX_SIZE] __attribute__((aligned(16))); /* << 19 */
	uint32_t max_exponent[MAX_CHANNELS];
#endif
	int16_t coefs1[MAX_CHANNELS][BLOCK_MAX_SIZE];
	int32_t coefs[MAX_CHANNELS][BLOCK_MAX_SIZE] __attribute__((aligned(16)));
	int32_t *windows[BLOCK_NB_SIZES];
	/* output buffer for one frame and the last for IMDCT windowing */
	int32_t frame_out[MAX_CHANNELS][BLOCK_MAX_SIZE * 2] __attribute__((aligned(16)));
	/* last frame info */
	uint8_t last_superframe[MAX_CODED_SUPERFRAME_SIZE + 4]; /* padding added */
	int last_bitoffset;
	int last_superframe_len;
#ifdef USEFLOAT
	float noise_table[NOISE_TAB_SIZE];
#else
	int32_t noise_table[NOISE_TAB_SIZE];       /* << 16 */
#endif
	int noise_index;
	uint32_t noise_mult; /* XXX: suppress that and integrate it in the noise array */
	/* lsp_to_curve tables */
#ifdef USEFLOAT
	float lsp_cos_table[BLOCK_MAX_SIZE];
	float lsp_pow_e_table[256];
	float lsp_pow_m_table1[(1 << LSP_POW_BITS)];
	float lsp_pow_m_table2[(1 << LSP_POW_BITS)];
#else
	int32_t lsp_cos_table[BLOCK_MAX_SIZE];
#endif
	/* exp vlc table */
#ifdef USEFLOAT
	float pow_10_to_yover16[EXP_VLC_TAB_SIZE];
#else
	uint32_t pow_10_to_yover16[EXP_VLC_TAB_SIZE];
	uint32_t pow_10_to_yover16_rev[EXP_VLC_TAB_SIZE];
#endif
} WMADecodeContext;

int use_noise_coding;
int use_exp_vlc;

typedef struct CoefVLCTable {
	int n; /* total number of codes */
	const uint32_t *huffcodes; /* VLC bit values */
	const uint8_t *huffbits;   /* VLC bit size */
	const uint16_t *levels; /* table to build run/level tables */
} CoefVLCTable;

static void wma_lsp_to_curve_init(WMADecodeContext *s, int frame_len);
static void wma_exp_vlc_init(WMADecodeContext *s);

#include "wmadata.h"

/* XXX: use same run/length optimization as mpeg decoders */
static void init_coef_vlc(VLC *vlc,
						  uint16_t **prun_table, uint16_t **plevel_table,
						  const CoefVLCTable *vlc_table)
{
	int n = vlc_table->n;
	const uint8_t *table_bits = vlc_table->huffbits;
	const uint32_t *table_codes = vlc_table->huffcodes;
	const uint16_t *levels_table = vlc_table->levels;
	uint16_t *run_table, *level_table;
	const uint16_t *p;
	int i, l, j, level;

	init_vlc(vlc, 9, n, table_bits, 1, 1, table_codes, 4, 4);

	run_table = (uint16_t *)malloc(n * sizeof(uint16_t));
	level_table = (uint16_t *)malloc(n * sizeof(uint16_t));
	p = levels_table;
	i = 2;
	level = 1;
	while (i < n) {
		l = *p++;
		for(j=0;j<l;j++) {
			run_table[i] = j;
			level_table[i] = level;
			i++;
		}
		level++;
	}
	*prun_table = run_table;
	*plevel_table = level_table;
}

void *wma_decode_init(void)
{
	WMADecodeContext *s;
	int i, flags1, flags2;
	int32_t *window;
	//uint8_t *extradata;
	float bps1, high_freq;
	volatile float bps;
	int sample_rate1;
	int coef_vlc_table;

	if(audheader == NULL) return NULL;
	if((audheader->wFormatTag != 0x0160) && (audheader->wFormatTag != 0x0161)) return NULL;

	s = (WMADecodeContext *)malloc(sizeof(WMADecodeContext));
	if(s == NULL) return NULL;

	s->sample_rate = audheader->nSamplesPerSec;
	s->nb_channels = audheader->nChannels;
	s->bit_rate = audheader->nAvgBytesPerSec * 8;
	s->block_align = audheader->nBlockAlign;
	s->version = (audheader->wFormatTag == 0x0160) ? 1 : 2;

	/* extract flag infos */
	flags1 = 0;
	flags2 = 0;

	if (s->version == 1 && audextsize >= 4) {
		flags1 = audextdata[0] | (audextdata[1] << 8);
		flags2 = audextdata[2] | (audextdata[3] << 8);
	} else if (s->version == 2 && audextsize >= 6) {
		flags1 = audextdata[0] | (audextdata[1] << 8) |
			(audextdata[2] << 16) | (audextdata[3] << 24);
		flags2 = audextdata[4] | (audextdata[5] << 8);
	}

	s->use_exp_vlc = flags2 & 0x0001;
	s->use_bit_reservoir = flags2 & 0x0002;
	s->use_variable_block_len = flags2 & 0x0004;

	/* compute MDCT block size */
	if (s->sample_rate <= 16000) {
		s->frame_len_bits = 9;
	} else if (s->sample_rate <= 22050 ||
			   (s->sample_rate <= 32000 && s->version == 1)) {
		s->frame_len_bits = 10;
	} else {
		s->frame_len_bits = 11;
	}


	s->frame_len = 1 << s->frame_len_bits;
	if (s->use_variable_block_len) {
		int nb_max, nb;
		nb = ((flags2 >> 3) & 3) + 1;
		if ((s->bit_rate / s->nb_channels) >= 32000)
			nb += 2;
		nb_max = s->frame_len_bits - BLOCK_MIN_BITS;
		if (nb > nb_max)
			nb = nb_max;
		s->nb_block_sizes = nb + 1;
	} else {
		s->nb_block_sizes = 1;
	}

	/* init rate dependant parameters */
	s->use_noise_coding = 1;
	high_freq = s->sample_rate * 0.5;

	/* if version 2, then the rates are normalized */
	sample_rate1 = s->sample_rate;
	if (s->version == 2) {
		if (sample_rate1 >= 44100)
			sample_rate1 = 44100;
		else if (sample_rate1 >= 22050)
			sample_rate1 = 22050;
		else if (sample_rate1 >= 16000)
			sample_rate1 = 16000;
		else if (sample_rate1 >= 11025)
			sample_rate1 = 11025;
		else if (sample_rate1 >= 8000)
			sample_rate1 = 8000;
	}

	bps = (float)s->bit_rate / (float)(s->nb_channels * s->sample_rate);
	s->byte_offset_bits = av_log2((int)(bps * s->frame_len / 8.0)) + 2;

	/* compute high frequency value and choose if noise coding should
	   be activated */
	bps1 = bps;
	if (s->nb_channels == 2)
		bps1 = bps * 1.6;
	if (sample_rate1 == 44100) {
		if (bps1 >= 0.61)
			s->use_noise_coding = 0;
		else
			high_freq = high_freq * 0.4;
	} else if (sample_rate1 == 22050) {
		if (bps1 >= 1.16)
			s->use_noise_coding = 0;
		else if (bps1 >= 0.72)
			high_freq = high_freq * 0.7;
		else
			high_freq = high_freq * 0.6;
	} else if (sample_rate1 == 16000) {
		if (bps > 0.5)
			high_freq = high_freq * 0.5;
		else
			high_freq = high_freq * 0.3;
	} else if (sample_rate1 == 11025) {
		high_freq = high_freq * 0.7;
	} else if (sample_rate1 == 8000) {
		if (bps <= 0.625) {
			high_freq = high_freq * 0.5;
		} else if (bps > 0.75) {
			s->use_noise_coding = 0;
		} else {
			high_freq = high_freq * 0.65;
		}
	} else {
		if (bps >= 0.8) {
			high_freq = high_freq * 0.75;
		} else if (bps >= 0.6) {
			high_freq = high_freq * 0.6;
		} else {
			high_freq = high_freq * 0.5;
		}
	}

	dprintf("flags1=0x%x flags2=0x%x\n", flags1, flags2);
	dprintf("version=%d channels=%d sample_rate=%d bitrate=%d block_align=%d\n",
		   s->version, s->nb_channels, s->sample_rate, s->bit_rate,
		   s->block_align);
	dprintf("bps=%f bps1=%f high_freq=%f bitoffset=%d\n",
		   bps, bps1, high_freq, s->byte_offset_bits);
	dprintf("use_noise_coding=%d use_exp_vlc=%d nb_block_sizes=%d\n",
		   s->use_noise_coding, s->use_exp_vlc, s->nb_block_sizes);

	//printf("noise coding=%s, exponent coding=%s\n", s->use_noise_coding ? "Yes" : "No", s->use_exp_vlc ? "VLC" : "LSP");
	use_noise_coding = s->use_noise_coding;
	use_exp_vlc = s->use_exp_vlc;

	/* compute the scale factor band sizes for each MDCT block size */
	{
		int a, b, pos, lpos, k, block_len, i, j, n;
		const uint8_t *table;

		if (s->version == 1) {
			s->coefs_start = 3;
		} else {
			s->coefs_start = 0;
		}
		for(k = 0; k < s->nb_block_sizes; k++) {
			block_len = s->frame_len >> k;

			if (s->version == 1) {
				lpos = 0;
				for(i=0;i<25;i++) {
					a = wma_critical_freqs[i];
					b = s->sample_rate;
					pos = ((block_len * 2 * a)  + (b >> 1)) / b;
					if (pos > block_len)
						pos = block_len;
					s->exponent_bands[0][i] = pos - lpos;
					if (pos >= block_len) {
						i++;
						break;
					}
					lpos = pos;
				}
				s->exponent_sizes[0] = i;
			} else {
				/* hardcoded tables */
				table = NULL;
				a = s->frame_len_bits - BLOCK_MIN_BITS - k;
				if (a < 3) {
					if (s->sample_rate >= 44100)
						table = exponent_band_44100[a];
					else if (s->sample_rate >= 32000)
						table = exponent_band_32000[a];
					else if (s->sample_rate >= 22050)
						table = exponent_band_22050[a];
				}
				if (table) {
					n = *table++;
					for(i=0;i<n;i++)
						s->exponent_bands[k][i] = table[i];
					s->exponent_sizes[k] = n;
				} else {
					j = 0;
					lpos = 0;
					for(i=0;i<25;i++) {
						a = wma_critical_freqs[i];
						b = s->sample_rate;
						pos = ((block_len * 2 * a)  + (b << 1)) / (4 * b);
						pos <<= 2;
						if (pos > block_len)
							pos = block_len;
						if (pos > lpos)
							s->exponent_bands[k][j++] = pos - lpos;
						if (pos >= block_len)
							break;
						lpos = pos;
					}
					s->exponent_sizes[k] = j;
				}
			}

			/* max number of coefs */
			s->coefs_end[k] = (s->frame_len - ((s->frame_len * 9) / 100)) >> k;
			/* high freq computation */
			s->high_band_start[k] = (int)((block_len * 2 * high_freq) /
										  s->sample_rate + 0.5);
			n = s->exponent_sizes[k];
			j = 0;
			pos = 0;
			for(i=0;i<n;i++) {
				int start, end;
				start = pos;
				pos += s->exponent_bands[k][i];
				end = pos;
				if (start < s->high_band_start[k])
					start = s->high_band_start[k];
				if (end > s->coefs_end[k])
					end = s->coefs_end[k];
				if (end > start)
					s->exponent_high_bands[k][j++] = end - start;
			}
			s->exponent_high_sizes[k] = j;
#if 1
			dprintf("%5d: coefs_end=%d high_band_start=%d nb_high_bands=%d: ",
				  s->frame_len >> k,
				  s->coefs_end[k],
				  s->high_band_start[k],
				  s->exponent_high_sizes[k]);
			for(j=0;j<s->exponent_high_sizes[k];j++)
				dprintf(" %d", s->exponent_high_bands[k][j]);
			dprintf("\n");
#endif
		}
	}

	/* init MDCT windows : simple sinus window */
	for(i = 0; i < s->nb_block_sizes; i++) {
		int n, j;
		float alpha;
		n = 1 << (s->frame_len_bits - i);
		window = (int32_t *)malloc(sizeof(int32_t) * n);
		alpha = M_PI / (2.0 * n);
		for(j=0;j<n;j++) {
			window[n - j - 1] = NFPTOFXS(sin((j + 0.5) * alpha), 31);
		}
		s->windows[i] = window;
	}

	s->reset_block_lengths = 1;

	if (s->use_noise_coding) {

		/* init the noise generator */
		if (s->use_exp_vlc)
			s->noise_mult = 50; // 1 / 0.02
		else
			s->noise_mult = 25; // 1 / 0.04

		{
			unsigned int seed = 1;
			float norm = 886.81001347584 / s->noise_mult / (float)(1LL << 31); // 512 * sqrt(3)
			for(i=0;i<NOISE_TAB_SIZE;i++) {
				float val;
				seed = seed * 314159 + 1;
				val = (float)((int)seed) * norm;
#ifdef USEFLOAT
				s->noise_table[i] = val;
#else
				s->noise_table[i] = FPTOFXS(val, 16);
#endif
			}
		}

		init_vlc(&s->hgain_vlc, 9, sizeof(hgain_huffbits),
				 hgain_huffbits, 1, 1,
				 hgain_huffcodes, 2, 2);
	}
	if (s->use_exp_vlc) {
		wma_exp_vlc_init(s);
		init_vlc(&s->exp_vlc, 9, sizeof(scale_huffbits),
				 scale_huffbits, 1, 1,
				 scale_huffcodes, 4, 4);
	} else {
		wma_lsp_to_curve_init(s, s->frame_len);
	}

	/* choose the VLC tables for the coefficients */
	coef_vlc_table = 2;
	if (s->sample_rate >= 32000) {
		if (bps1 < 0.72)
			coef_vlc_table = 0;
		else if (bps1 < 1.16)
			coef_vlc_table = 1;
	}

	init_coef_vlc(&s->coef_vlc[0], &s->run_table[0], &s->level_table[0],
				  &coef_vlcs[coef_vlc_table * 2]);
	init_coef_vlc(&s->coef_vlc[1], &s->run_table[1], &s->level_table[1],
				  &coef_vlcs[coef_vlc_table * 2 + 1]);
	return s;
}

/* interpolate values for a bigger or smaller block. The block must
   have multiple sizes */
#ifdef USEFLOAT
static void interpolate_array(float *scale, int old_size, int new_size)
#else
static void interpolate_array(uint32_t *scale, int old_size, int new_size)
#endif
{
	int i, j, jincr, k;
#ifdef USEFLOAT
	float v;
#else
	uint32_t v;
#endif

	if (new_size > old_size) {
		jincr = new_size / old_size;
		j = new_size;
		for(i = old_size - 1; i >=0; i--) {
			v = scale[i];
			k = jincr;
			do {
				scale[--j] = v;
			} while (--k);
		}
	} else if (new_size < old_size) {
		j = 0;
		jincr = old_size / new_size;
		for(i = 0; i < new_size; i++) {
			scale[i] = scale[j];
			j += jincr;
		}
	}
}

#ifdef USEFLOAT
/* compute x^-0.25 with an exponent and mantissa table. We use linear
   interpolation to reduce the mantissa table size at a small speed
   expense (linear interpolation approximately doubles the number of
   bits of precision). */
static inline float pow_m1_4(WMADecodeContext *s, float x)
{
	union {
		float f;
		unsigned int v;
	} u, t;
	unsigned int e, m;
	float a, b;

	u.f = x;
	e = u.v >> 23;
	m = (u.v >> (23 - LSP_POW_BITS)) & ((1 << LSP_POW_BITS) - 1);
	/* build interpolation scale: 1 <= t < 2. */
	t.v = ((u.v << LSP_POW_BITS) & ((1 << 23) - 1)) | (127 << 23);
	a = s->lsp_pow_m_table1[m];
	b = s->lsp_pow_m_table2[m];
	return s->lsp_pow_e_table[e] * (a + b * t.f);
}
#endif

#ifndef USEFLOAT
#define COS_LOOKUP_I_SHIFT 9
#define COS_LOOKUP_I_MASK 511
#define COS_LOOKUP_I_SZ 128
static const int32_t COS_LOOKUP_I[COS_LOOKUP_I_SZ+1]={
	   16384,   16379,   16364,   16340,
	   16305,   16261,   16207,   16143,
	   16069,   15986,   15893,   15791,
	   15679,   15557,   15426,   15286,
	   15137,   14978,   14811,   14635,
	   14449,   14256,   14053,   13842,
	   13623,   13395,   13160,   12916,
	   12665,   12406,   12140,   11866,
	   11585,   11297,   11003,   10702,
	   10394,   10080,    9760,    9434,
	    9102,    8765,    8423,    8076,
	    7723,    7366,    7005,    6639,
	    6270,    5897,    5520,    5139,
	    4756,    4370,    3981,    3590,
	    3196,    2801,    2404,    2006,
	    1606,    1205,     804,     402,
	       0,    -401,    -803,   -1204,
	   -1605,   -2005,   -2403,   -2800,
	   -3195,   -3589,   -3980,   -4369,
	   -4755,   -5138,   -5519,   -5896,
	   -6269,   -6638,   -7004,   -7365,
	   -7722,   -8075,   -8422,   -8764,
	   -9101,   -9433,   -9759,  -10079,
	  -10393,  -10701,  -11002,  -11296,
	  -11584,  -11865,  -12139,  -12405,
	  -12664,  -12915,  -13159,  -13394,
	  -13622,  -13841,  -14052,  -14255,
	  -14448,  -14634,  -14810,  -14977,
	  -15136,  -15285,  -15425,  -15556,
	  -15678,  -15790,  -15892,  -15985,
	  -16068,  -16142,  -16206,  -16260,
	  -16304,  -16339,  -16363,  -16378,
	  -16383,
};
#endif

static void wma_lsp_to_curve_init(WMADecodeContext *s, int frame_len)
{
	int i;
#ifdef USEFLOAT
	const float wdel = M_PI / frame_len;
	float a, b;
	int m;
#else
	//const float wdelf = M_PI / frame_len;
	const uint32_t wdel = 65535 / frame_len;
	uint32_t a, n, d;
#endif

	for(i=0;i<frame_len;i++) {
#ifdef USEFLOAT
		s->lsp_cos_table[i] = 2.0f * cos(wdel * i);
#else
		/* interpolated lookup based cos function, domain 0 to PI only */
		/* a is in 0.16 format, where 0==0, 2^^16-1==PI, return 0.14 */
		a = wdel * i;
		n = a >> COS_LOOKUP_I_SHIFT;
		d = a & COS_LOOKUP_I_MASK;
		s->lsp_cos_table[i] = COS_LOOKUP_I[n] - ((d*(COS_LOOKUP_I[n]-COS_LOOKUP_I[n+1]))>>COS_LOOKUP_I_SHIFT);
		s->lsp_cos_table[i] <<= 6;
		//printf("%d/%d, %.8f, %.8f\n", i, frame_len, s->lsp_cos_table[i] / 524287.0f, 2.0f * cos(wdelf * i));
#endif
	}

#ifdef USEFLOAT
    /* NOTE: these two tables are needed to avoid two operations in
       pow_m1_4 */
    b = 1.0;
    for(i=(1 << LSP_POW_BITS) - 1;i>=0;i--) {
        m = (1 << LSP_POW_BITS) + i;
        a = (float)m * (0.5 / (1 << LSP_POW_BITS));
        a = pow(a, -0.25);
        s->lsp_pow_m_table1[i] = 2 * a - b;
        s->lsp_pow_m_table2[i] = b - a;
        b = a;
    }
#  if 0
	for(i=1;i<20;i++) {
		float v, r1, r2;
		v = 5.0 / i;
		r1 = pow_m1_4(s, v);
		r2 = pow(v,-0.25);
		printf("%f^-0.25=%f e=%f\n", v, r1, r2 - r1);
	}
#  endif
#else
#  if 0
	for(i=0;i<0x1000;i++) {
		lsp_pow_table[0][i] = FPTOFXU(pow(i, -0.25), 24);
		lsp_pow_table[1][i] = FPTOFXU(pow(i, 0.25), 19);
	}

	for(i=0;i<0x1000;i++) {
		if(!(i % 10)) printf("\n");
		printf("%u, ", lsp_pow_table[0][i]);
	}
	printf("\n\n");
	for(i=0;i<0x1000;i++) {
		if(!(i % 10)) printf("\n");
		printf("%u, ", lsp_pow_table[1][i]);
	}
#  endif
#endif // USEFLOAT
}

/* NOTE: We use the same code as Vorbis here */
/* XXX: optimize it further with SSE/3Dnow */
static void wma_lsp_to_curve(WMADecodeContext *s,
#ifdef USEFLOAT
							 float *out, float *val_max_ptr, int n, float *lsp)
#else
							 uint32_t *out, uint32_t *val_max_ptr, int n, int32_t *lsp)
#endif
{
	int i, j;
#ifdef USEFLOAT
	float p, q, w, v, val_max;
#else
	int64_t p, q;
	int32_t w, v;
	uint32_t x, y, r;
	uint32_t val_cmp = 0, val_max;
#endif

	val_max = 0;
	dprintf("use wma_lsp_to_curve...\n");
	for(i=0;i<n;i++) {
#ifdef USEFLOAT
		p = 0.5f;
		q = 0.5f;
		w = s->lsp_cos_table[i];
		for(j=1;j<NB_LSP_COEFS;j+=2){
			q *= w - lsp[j - 1];
			p *= w - lsp[j];
		}
		p *= p * (2.0f - w);
		q *= q * (2.0f + w);
		v = p + q;
		v = pow_m1_4(s, v);
		if (v > val_max)
			val_max = v;
		out[i] = v;
#else
		p = 524288;
		q = 524288;
		w = s->lsp_cos_table[i];
		for(j=1;j<NB_LSP_COEFS;j+=2){
			q = IRINT(q * (w - lsp[j - 1]), 19);
			p = IRINT(p * (w - lsp[j]), 19);
		}
		p = IRINT(IRINT(p * p, 20) * (1048574 - w), 19);
		q = IRINT(IRINT(q * q, 20) * (1048574 + w), 19);
		v = p + q;

		x = v;
		y = 0;
		while((x & 0xfff) != x) { x >>= 4; y++; }
		r = lsp_pow_table[0][x] >> y;
		if(r > val_cmp) {
			val_cmp = r;
			val_max = lsp_pow_table[1][x] >> (5 - y);
		}
		out[i] = r;
#endif
	}
#ifdef USEFLOAT
	*val_max_ptr = 1 / val_max;
#else
	*val_max_ptr = val_max;
#endif
}

/* decode exponents coded with LSP coefficients (same idea as Vorbis) */
static void decode_exp_lsp(WMADecodeContext *s, int ch)
{
#ifdef USEFLOAT
	float lsp_coefs[NB_LSP_COEFS];
#else
	int32_t lsp_coefs[NB_LSP_COEFS];
#endif
	int val, i;

	dprintf("decode_exp_lsp\n");
	for(i = 0; i < NB_LSP_COEFS; i++) {
		//if ((i & 0x1) == 0) bitload(&s->gb);
		if (i == 0 || i >= 8)
			val = get_bits(&s->gb, 3);
		else
			val = get_bits(&s->gb, 4);
		lsp_coefs[i] = lsp_codebook[i][val];
	}
/*	
	{
		static int flag = 1;
		if(flag) {
			int i, j;
			for(i = 0; i < NB_LSP_COEFS; i++) {
				for(j = 0; j < 16; j++) {
					float val = lsp_codebook[i][j];
					int32_t v = FPTOFXS(val, 19);
					if(!(j % 8)) printf("\n");
					printf("%d, ", v);
				}
				printf("\n\n");
			}
			
			flag = 0;
		}
	}
*/
	wma_lsp_to_curve(s, s->exponents[ch], &s->max_exponent[ch], s->block_len, lsp_coefs);
	CONDTRIG("decode_exp_lsp");
}

static void wma_exp_vlc_init(WMADecodeContext *s)
{
	/* replaced code with pre calculated values */
#ifdef USEFLOAT
	s->pow_10_to_yover16[0] = 1.000000;
	s->pow_10_to_yover16[1] = 1.154782;
	s->pow_10_to_yover16[2] = 1.333521;
	s->pow_10_to_yover16[3] = 1.539927;
	s->pow_10_to_yover16[4] = 1.778279;
	s->pow_10_to_yover16[5] = 2.053525;
	s->pow_10_to_yover16[6] = 2.371374;
	s->pow_10_to_yover16[7] = 2.738420;
	s->pow_10_to_yover16[8] = 3.162278;
	s->pow_10_to_yover16[9] = 3.651741;
	s->pow_10_to_yover16[10] = 4.216965;
	s->pow_10_to_yover16[11] = 4.869675;
	s->pow_10_to_yover16[12] = 5.623413;
	s->pow_10_to_yover16[13] = 6.493816;
	s->pow_10_to_yover16[14] = 7.498942;
	s->pow_10_to_yover16[15] = 8.659643;
	s->pow_10_to_yover16[16] = 10.000000;
	s->pow_10_to_yover16[17] = 11.547820;
	s->pow_10_to_yover16[18] = 13.335214;
	s->pow_10_to_yover16[19] = 15.399265;
	s->pow_10_to_yover16[20] = 17.782794;
	s->pow_10_to_yover16[21] = 20.535250;
	s->pow_10_to_yover16[22] = 23.713737;
	s->pow_10_to_yover16[23] = 27.384196;
	s->pow_10_to_yover16[24] = 31.622777;
	s->pow_10_to_yover16[25] = 36.517413;
	s->pow_10_to_yover16[26] = 42.169650;
	s->pow_10_to_yover16[27] = 48.696753;
	s->pow_10_to_yover16[28] = 56.234133;
	s->pow_10_to_yover16[29] = 64.938163;
	s->pow_10_to_yover16[30] = 74.989421;
	s->pow_10_to_yover16[31] = 86.596432;
	s->pow_10_to_yover16[32] = 100.000000;
	s->pow_10_to_yover16[33] = 115.478198;
	s->pow_10_to_yover16[34] = 133.352143;
	s->pow_10_to_yover16[35] = 153.992653;
	s->pow_10_to_yover16[36] = 177.827941;
	s->pow_10_to_yover16[37] = 205.352503;
	s->pow_10_to_yover16[38] = 237.137371;
	s->pow_10_to_yover16[39] = 273.841963;
	s->pow_10_to_yover16[40] = 316.227766;
	s->pow_10_to_yover16[41] = 365.174127;
	s->pow_10_to_yover16[42] = 421.696503;
	s->pow_10_to_yover16[43] = 486.967525;
	s->pow_10_to_yover16[44] = 562.341325;
	s->pow_10_to_yover16[45] = 649.381632;
	s->pow_10_to_yover16[46] = 749.894209;
	s->pow_10_to_yover16[47] = 865.964323;
	s->pow_10_to_yover16[48] = 1000.000000;
	s->pow_10_to_yover16[49] = 1154.781985;
	s->pow_10_to_yover16[50] = 1333.521432;
	s->pow_10_to_yover16[51] = 1539.926526;
	s->pow_10_to_yover16[52] = 1778.279410;
	s->pow_10_to_yover16[53] = 2053.525026;
	s->pow_10_to_yover16[54] = 2371.373706;
	s->pow_10_to_yover16[55] = 2738.419634;
	s->pow_10_to_yover16[56] = 3162.277660;
	s->pow_10_to_yover16[57] = 3651.741273;
	s->pow_10_to_yover16[58] = 4216.965034;
	s->pow_10_to_yover16[59] = 4869.675252;
	s->pow_10_to_yover16[60] = 5623.413252;
	s->pow_10_to_yover16[61] = 6493.816316;
	s->pow_10_to_yover16[62] = 7498.942093;
	/*
	{
		int i;
		for(i=0;i<63;i++) {
			float val = 1/s->pow_10_to_yover16[i];
			printf("\ts->pow_10_to_yover16_rev[%d] = %ud\n", i, FPTOFXU(val, 19));
		}
	}
	*/
#else
	s->pow_10_to_yover16[0] = 524287;
	s->pow_10_to_yover16[1] = 605437;
	s->pow_10_to_yover16[2] = 699148;
	s->pow_10_to_yover16[3] = 807364;
	s->pow_10_to_yover16[4] = 932329;
	s->pow_10_to_yover16[5] = 1076636;
	s->pow_10_to_yover16[6] = 1243280;
	s->pow_10_to_yover16[7] = 1435718;
	s->pow_10_to_yover16[8] = 1657941;
	s->pow_10_to_yover16[9] = 1914560;
	s->pow_10_to_yover16[10] = 2210900;
	s->pow_10_to_yover16[11] = 2553108;
	s->pow_10_to_yover16[12] = 2948282;
	s->pow_10_to_yover16[13] = 3404623;
	s->pow_10_to_yover16[14] = 3931598;
	s->pow_10_to_yover16[15] = 4540138;
	s->pow_10_to_yover16[16] = 5242870;
	s->pow_10_to_yover16[17] = 6054372;
	s->pow_10_to_yover16[18] = 6991479;
	s->pow_10_to_yover16[19] = 8073634;
	s->pow_10_to_yover16[20] = 9323288;
	s->pow_10_to_yover16[21] = 10766364;
	s->pow_10_to_yover16[22] = 12432804;
	s->pow_10_to_yover16[23] = 14357178;
	s->pow_10_to_yover16[24] = 16579411;
	s->pow_10_to_yover16[25] = 19145606;
	s->pow_10_to_yover16[26] = 22109000;
	s->pow_10_to_yover16[27] = 25531076;
	s->pow_10_to_yover16[28] = 29482826;
	s->pow_10_to_yover16[29] = 34046236;
	s->pow_10_to_yover16[30] = 39315976;
	s->pow_10_to_yover16[31] = 45401384;
	s->pow_10_to_yover16[32] = 52428700;
	s->pow_10_to_yover16[33] = 60543716;
	s->pow_10_to_yover16[34] = 69914792;
	s->pow_10_to_yover16[35] = 80736352;
	s->pow_10_to_yover16[36] = 93232880;
	s->pow_10_to_yover16[37] = 107663648;
	s->pow_10_to_yover16[38] = 124328040;
	s->pow_10_to_yover16[39] = 143571776;
	s->pow_10_to_yover16[40] = 165794096;
	s->pow_10_to_yover16[41] = 191456048;
	s->pow_10_to_yover16[42] = 221090000;
	s->pow_10_to_yover16[43] = 255310752;
	s->pow_10_to_yover16[44] = 294828224;
	s->pow_10_to_yover16[45] = 340462368;
	s->pow_10_to_yover16[46] = 393159808;
	s->pow_10_to_yover16[47] = 454013824;
	s->pow_10_to_yover16[48] = 524287008;
	s->pow_10_to_yover16[49] = 605437184;
	s->pow_10_to_yover16[50] = 699147968;
	s->pow_10_to_yover16[51] = 807363456;
	s->pow_10_to_yover16[52] = 932328768;
	s->pow_10_to_yover16[53] = 1076636544;
	s->pow_10_to_yover16[54] = 1243280384;
	s->pow_10_to_yover16[55] = 1435717888;
	s->pow_10_to_yover16[56] = 1657940992;
	s->pow_10_to_yover16[57] = 1914560384;
	s->pow_10_to_yover16[58] = 2210899968UL;
	s->pow_10_to_yover16[59] = 2553107456UL;
	s->pow_10_to_yover16[60] = 2948282368UL;
	s->pow_10_to_yover16[61] = 3404623616UL;
	s->pow_10_to_yover16[62] = 3931597824UL;

	s->pow_10_to_yover16_rev[0] = 524287;
	s->pow_10_to_yover16_rev[1] = 454014;
	s->pow_10_to_yover16_rev[2] = 393160;
	s->pow_10_to_yover16_rev[3] = 340462;
	s->pow_10_to_yover16_rev[4] = 294828;
	s->pow_10_to_yover16_rev[5] = 255311;
	s->pow_10_to_yover16_rev[6] = 221090;
	s->pow_10_to_yover16_rev[7] = 191456;
	s->pow_10_to_yover16_rev[8] = 165794;
	s->pow_10_to_yover16_rev[9] = 143572;
	s->pow_10_to_yover16_rev[10] = 124328;
	s->pow_10_to_yover16_rev[11] = 107664;
	s->pow_10_to_yover16_rev[12] = 93233;
	s->pow_10_to_yover16_rev[13] = 80736;
	s->pow_10_to_yover16_rev[14] = 69915;
	s->pow_10_to_yover16_rev[15] = 60544;
	s->pow_10_to_yover16_rev[16] = 52429;
	s->pow_10_to_yover16_rev[17] = 45401;
	s->pow_10_to_yover16_rev[18] = 39316;
	s->pow_10_to_yover16_rev[19] = 34046;
	s->pow_10_to_yover16_rev[20] = 29483;
	s->pow_10_to_yover16_rev[21] = 25531;
	s->pow_10_to_yover16_rev[22] = 22109;
	s->pow_10_to_yover16_rev[23] = 19146;
	s->pow_10_to_yover16_rev[24] = 16579;
	s->pow_10_to_yover16_rev[25] = 14357;
	s->pow_10_to_yover16_rev[26] = 12433;
	s->pow_10_to_yover16_rev[27] = 10766;
	s->pow_10_to_yover16_rev[28] = 9323;
	s->pow_10_to_yover16_rev[29] = 8074;
	s->pow_10_to_yover16_rev[30] = 6991;
	s->pow_10_to_yover16_rev[31] = 6054;
	s->pow_10_to_yover16_rev[32] = 5243;
	s->pow_10_to_yover16_rev[33] = 4540;
	s->pow_10_to_yover16_rev[34] = 3932;
	s->pow_10_to_yover16_rev[35] = 3405;
	s->pow_10_to_yover16_rev[36] = 2948;
	s->pow_10_to_yover16_rev[37] = 2553;
	s->pow_10_to_yover16_rev[38] = 2211;
	s->pow_10_to_yover16_rev[39] = 1915;
	s->pow_10_to_yover16_rev[40] = 1658;
	s->pow_10_to_yover16_rev[41] = 1436;
	s->pow_10_to_yover16_rev[42] = 1243;
	s->pow_10_to_yover16_rev[43] = 1077;
	s->pow_10_to_yover16_rev[44] = 932;
	s->pow_10_to_yover16_rev[45] = 807;
	s->pow_10_to_yover16_rev[46] = 699;
	s->pow_10_to_yover16_rev[47] = 605;
	s->pow_10_to_yover16_rev[48] = 524;
	s->pow_10_to_yover16_rev[49] = 454;
	s->pow_10_to_yover16_rev[50] = 393;
	s->pow_10_to_yover16_rev[51] = 340;
	s->pow_10_to_yover16_rev[52] = 295;
	s->pow_10_to_yover16_rev[53] = 255;
	s->pow_10_to_yover16_rev[54] = 221;
	s->pow_10_to_yover16_rev[55] = 191;
	s->pow_10_to_yover16_rev[56] = 166;
	s->pow_10_to_yover16_rev[57] = 144;
	s->pow_10_to_yover16_rev[58] = 124;
	s->pow_10_to_yover16_rev[59] = 108;
	s->pow_10_to_yover16_rev[60] = 93;
	s->pow_10_to_yover16_rev[61] = 81;
	s->pow_10_to_yover16_rev[62] = 70;
#endif
}

/* decode exponents coded with VLC codes */
static int decode_exp_vlc(WMADecodeContext *s, int ch)
{
	int last_exp, n, code;
	const uint16_t *ptr, *band_ptr;
#ifdef USEFLOAT
	float v, *q, *q_end, max_scale;
#else
	uint32_t v, *q, *q_end, max_scale, max_index = 0;
#endif

	band_ptr = s->exponent_bands[s->frame_len_bits - s->block_len_bits];
	ptr = band_ptr;
	q = s->exponents[ch];
	q_end = q + s->block_len;
	max_scale = 0;
	if (s->version == 1) {
		bitload(&s->gb);
		last_exp = get_bits(&s->gb, 5) + 10;
		//assert((last_exp >= 0) && (last_exp < EXP_VLC_TAB_SIZE));
		v = s->pow_10_to_yover16[last_exp];
		max_scale = v;
#ifndef USEFLOAT
		max_index = last_exp;
#endif
		n = *ptr++;
		do {
			*q++ = v;
		} while (--n);
	}
	last_exp = 36;
	while (q < q_end) {
		code = get_vlc(&s->gb, &s->exp_vlc);
		if (code < 0)
			return -1;
		/* NOTE: this offset is the same as MPEG4 AAC ! */
		last_exp += code - 60;

		if(last_exp >= EXP_VLC_TAB_SIZE) last_exp = EXP_VLC_TAB_SIZE-1;
		if(last_exp < 0) last_exp = 0;

		//assert((last_exp >= 0) && (last_exp < EXP_VLC_TAB_SIZE));
		v = s->pow_10_to_yover16[last_exp];
		if (v > max_scale) {
			max_scale = v;
#ifndef USEFLOAT
			max_index = last_exp;
#endif
		}
		n = *ptr++;
		do {
			*q++ = v;
		} while (--n);
	}
#ifdef USEFLOAT
	s->max_exponent[ch] = 1 / max_scale;
#else
	s->max_exponent[ch] = s->pow_10_to_yover16_rev[max_index];
#endif
	return 0;
}


/* The following is a set of precalculated results for the calculation
** of ten to the power of (int multiplied by 0.05)
** This needs to be tidied up
*/

#ifdef USEFLOAT
static const float internal_05_power_ten_table[200] = {
	1.000000, 1.122018, 1.258925, 1.412538, 1.584893, 1.778279, 1.995262, 2.238721, 2.511886, 2.818383,
	3.162278, 3.548134, 3.981072, 4.466836, 5.011872, 5.623413, 6.309574, 7.079458, 7.943282, 8.912509,
	10.000000, 11.220184, 12.589254, 14.125376, 15.848932, 17.782795, 19.952623, 22.387211, 25.118864, 28.183830,
	31.622776, 35.481339, 39.810719, 44.668358, 50.118725, 56.234131, 63.095734, 70.794579, 79.432823, 89.125092,
	100.000000, 112.201843, 125.892540, 141.253754, 158.489319, 177.827942, 199.526230, 223.872116, 251.188644, 281.838287,
	316.227753, 354.813385, 398.107178, 446.683594, 501.187225, 562.341309, 630.957336, 707.945801, 794.328247, 891.250916,
	1000.000000, 1122.018433, 1258.925415, 1412.537598, 1584.893188, 1778.279419, 1995.262329, 2238.721191, 2511.886475, 2818.382812,
	3162.277588, 3548.133789, 3981.071777, 4466.835938, 5011.872559, 5623.413086, 6309.573242, 7079.458008, 7943.282227, 8912.509766,
	10000.000000, 11220.184570, 12589.253906, 14125.375000, 15848.931641, 17782.794922, 19952.623047, 22387.210938, 25118.865234, 28183.830078,
	31622.777344, 35481.339844, 39810.718750, 44668.359375, 50118.722656, 56234.132812, 63095.734375, 70794.578125, 79432.820312, 89125.093750,
	100000.000000, 112201.843750, 125892.539062, 141253.750000, 158489.312500, 177827.937500, 199526.234375, 223872.109375, 251188.640625, 281838.281250,
	316227.781250, 354813.375000, 398107.156250, 446683.593750, 501187.218750, 562341.312500, 630957.375000, 707945.812500, 794328.250000, 891250.937500,
	1000000.000000, 1122018.500000, 1258925.375000, 1412537.500000, 1584893.250000, 1778279.375000, 1995262.375000, 2238721.250000, 2511886.500000, 2818383.000000,
	3162277.750000, 3548134.000000, 3981071.750000, 4466836.000000, 5011872.500000, 5623413.500000, 6309573.500000, 7079458.000000, 7943282.500000, 8912509.000000,
	10000000.000000, 11220185.000000, 12589254.000000, 14125375.000000, 15848932.000000, 17782794.000000, 19952624.000000, 22387212.000000, 25118864.000000, 28183830.000000,
	31622776.000000, 35481340.000000, 39810716.000000, 44668360.000000, 50118724.000000, 56234132.000000, 63095736.000000, 70794576.000000, 79432824.000000, 89125096.000000,
	100000000.000000, 112201848.000000, 125892544.000000, 141253760.000000, 158489312.000000, 177827936.000000, 199526224.000000, 223872112.000000, 251188640.000000, 281838304.000000,
	316227776.000000, 354813376.000000, 398107168.000000, 446683584.000000, 501187232.000000, 562341312.000000, 630957376.000000, 707945792.000000, 794328256.000000, 891250944.000000,
	1000000000.000000, 1122018432.000000, 1258925440.000000, 1412537600.000000, 1584893184.000000, 1778279424.000000, 1995262336.000000, 2238721024.000000, 2511886336.000000, 2818382848.000000,
	3162277632.000000, 3548133888.000000, 3981071616.000000, 4466835968.000000, 5011872256.000000, 5623413248.000000, 6309573632.000000, 7079457792.000000, 7943282176.000000, 8912508928.000000,
};
#else
static const uint64_t internal_05_power_ten_table[200] = {
	0x7ffLL, 0x8f8LL, 0xa11LL, 0xb4bLL, 0xcacLL, 0xe38LL, 0xff4LL, 0x11e6LL, 0x1415LL, 0x1689LL,
	0x1949LL, 0x1c5fLL, 0x1fd5LL, 0x23b7LL, 0x2813LL, 0x2cf7LL, 0x3273LL, 0x389bLL, 0x3f83LL, 0x4743LL,
	0x4ff6LL, 0x59b7LL, 0x64aaLL, 0x70f2LL, 0x7ebaLL, 0x8e31LL, 0x9f8bLL, 0xb302LL, 0xc8daLL, 0xe15cLL,
	0xfcdbLL, 0x11bb6LL, 0x13e54LL, 0x1652cLL, 0x190c1LL, 0x1c1a7LL, 0x1f884LL, 0x23614LL, 0x27b26LL, 0x2c8a7LL,
	0x31f9cLL, 0x3812dLL, 0x3eea6LL, 0x4697aLL, 0x4f34bLL, 0x58dedLL, 0x63b6eLL, 0x6fe1aLL, 0x7d887LL, 0x8cd9bLL,
	0x9e096LL, 0xb151fLL, 0xc6f4dLL, 0xdf3b9LL, 0xfa78aLL, 0x119088LL, 0x13b531LL, 0x161ccdLL, 0x18cf85LL, 0x1bd686LL,
	0x1f3c18LL, 0x230bc3LL, 0x27527cLL, 0x2c1ec8LL, 0x3180f4LL, 0x378b4aLL, 0x3e524eLL, 0x45ed06LL, 0x4e7547LL, 0x58080dLL,
	0x62c5deLL, 0x6ed336LL, 0x7c5906LL, 0x8b853dLL, 0x9c8b67LL, 0xafa557LL, 0xc513f0LL, 0xdd2003LL, 0xf81b3bLL, 0x1166144LL,
	0x13858f0LL, 0x15e75a6LL, 0x18938daLL, 0x1b933d2LL, 0x1ef098cLL, 0x22b70e4LL, 0x26f370cLL, 0x2bb423cLL, 0x31094ccLL, 0x370508cLL,
	0x3dbbab0LL, 0x4544020LL, 0x4db7a40LL, 0x5733460LL, 0x61d7200LL, 0x6dc7568LL, 0x7b2c768LL, 0x8a34010LL, 0x9b11040LL, 0xadfcca0LL,
	0xc337960LL, 0xdb09870LL, 0xf5c3890LL, 0x113c0640LL, 0x13565f60LL, 0x15b268e0LL, 0x18582680LL, 0x1b509660LL, 0x1ea5cfe0LL, 0x22632540LL,
	0x26954b00LL, 0x2b4a8100LL, 0x3092c640LL, 0x36800c00LL, 0x3d267400LL, 0x449c9600LL, 0x4cfbca80LL, 0x56608100LL, 0x60eaa300LL, 0x6cbdfe00LL,
	0x7a02be00LL, 0x88e5f500LL, 0x999a3500LL, 0xac583e00LL, 0xc15fbb00LL, 0xd8f81900LL, 0xf3718100LL, 0x11125e000LL, 0x1327a2000LL, 0x157df7600LL,
	0x181d4ee00LL, 0x1b0e90c00LL, 0x1e5bbbe00LL, 0x221007800LL, 0x263808c00LL, 0x2ae1ddc00LL, 0x31d5e400LL, 0x35fc50800LL, 0x3c92a6000LL, 0x43f6be800LL,
	0x4c41b6800LL, 0x558fb9000LL, 0x661800LL, 0x6bb727000LL, 0x78dbd4800LL, 0x879b0f000LL, 0x9826f1000LL, 0xaab7ac000LL, 0xbf8c53000LL, 0xd6ebaa000LL,
	0xf12514000LL, 0x10e91a8000LL, 0x12f9556000LL, 0x154a04a000LL, 0x17e3056000LL, 0x1acd2a8000LL, 0x1e125b0000LL, 0x21bdb24000LL, 0x25dba7c000LL, 0x2a7a374000LL,
	0x2fa9120000LL, 0x3579d3c000LL, 0x3c3d0000LL, 0x4352788000LL, 0x4b89648000LL, 0x54c0e98000LL, 0x5f18560000LL, 0x6ab2cb8000LL, 0x77b7b40000LL, 0x86534a0000LL,
	0x96b72d0000LL, 0xa91b080000LL, 0xbdbd560000LL, 0xd4e42e0000LL, 0xeede350000LL, 0x10c3aa0000LL, 0x12cb78e0000LL, 0x15168f80000LL, 0x17a948e0000LL, 0x1a8c6280000LL,
	0x1dc9ab60000LL, 0x216c2440000LL, 0x2582640000LL, 0x2a138b80000LL, 0x2f35df00000LL, 0x34f89200000LL, 0x3b6f3600000LL, 0x42afbf00000LL, 0x4ad2d080000LL, 0x53f40e00000LL,
	0x5e327b80000LL, 0x69be500000LL, 0x76965600000LL, 0x850e9d00000LL, 0x954ae100000LL, 0xa7824a00000LL, 0xbbf2b900000LL, 0xd2e19b00000LL, 0xec9cd800000LL, 0x1097bd800000LL
};
#endif

#ifdef DEBUG
#ifdef USEFLOAT
STIN float internal_05_power_ten(unsigned int gain)
#else
STIN uint64_t internal_05_power_ten(unsigned int gain)
#endif
{
	if(gain < 200) return internal_05_power_ten_table[gain];
	fprintf(stderr,"unhandled gain value passed %d\n", gain);
	return internal_05_power_ten_table[100];
}
#else
#define internal_05_power_ten(gain) internal_05_power_ten_table[(gain)]
#endif

/* return 0 if OK. return 1 if last block of frame. return -1 if
   unrecorrable error. */
static int wma_decode_block(WMADecodeContext *s)
{
	int n, v, a, ch, code, bsize;
	int coef_nb_bits, total_gain, parse_exponents;
	int32_t window[BLOCK_MAX_SIZE * 2];

#ifdef USEFLOAT
/*
	static int flag = 1;
	if(flag) {
		int i;
		for(i = 0; i < 200; i++) {
			uint64_t val = (uint64_t)(internal_05_power_ten(i) * ((1 << 11) - 1));
			uint32_t hi, lo;
			hi = val >> 32;
			lo = val & 0xffffffffUL;
			if(!(i % 10)) printf("\n");
			printf("0x");
			if(hi) printf("%x", hi);
			printf("%xLL, ", lo);
		}
		flag = 0;
	}
*/
#endif

// XXX: FIXME!! there's a bug somewhere which makes this mandatory under altivec
#ifdef HAVE_ALTIVEC
	volatile int nb_coefs[MAX_CHANNELS] __attribute__((aligned(16)));
	#error !!!
#else
	int nb_coefs[MAX_CHANNELS];
#endif
	int mdct_norm;

	/* compute current block length */
	if (s->use_variable_block_len) {
		n = av_log2(s->nb_block_sizes - 1) + 1;

		if (s->reset_block_lengths) {
			s->reset_block_lengths = 0;
					bitload(&s->gb);
			v = get_bits(&s->gb, n);
			if (v >= s->nb_block_sizes)
				return -1;
			s->prev_block_len_bits = s->frame_len_bits - v;
					bitload(&s->gb);
			v = get_bits(&s->gb, n);
			if (v >= s->nb_block_sizes)
				return -1;
			s->block_len_bits = s->frame_len_bits - v;
		} else {
			/* update block lengths */
			s->prev_block_len_bits = s->block_len_bits;
			s->block_len_bits = s->next_block_len_bits;
		}

			bitload(&s->gb);
		v = get_bits(&s->gb, n);
		if (v >= s->nb_block_sizes)
			return -1;
		s->next_block_len_bits = s->frame_len_bits - v;
	} else {

		/* fixed block len */
		s->next_block_len_bits = s->frame_len_bits;
		s->prev_block_len_bits = s->frame_len_bits;
		s->block_len_bits = s->frame_len_bits;
	}

	/* now check if the block length is coherent with the frame length */
	s->block_len = 1 << s->block_len_bits;
	if ((s->block_pos + s->block_len) > s->frame_len)
		return -1;

	bitload(&s->gb);
	if (s->nb_channels == 2) {
		s->ms_stereo = get_bits(&s->gb, 1);
	}
	v = 0;
	for(ch = 0; ch < s->nb_channels; ch++) {
		a = get_bits(&s->gb, 1);
		s->channel_coded[ch] = a;
		v |= a;
	}
	/* if no channel coded, no need to go further */
	/* XXX: fix potential framing problems */
	if (!v)
		goto next;

	bsize = s->frame_len_bits - s->block_len_bits;

	/* read total gain and extract corresponding number of bits for
	   coef escape coding */
	total_gain = 1;
	for(;;) {
			bitload(&s->gb);
		a = get_bits(&s->gb, 7);
		total_gain += a;
		if (a != 127)
			break;
	}

	if (total_gain < 15)
		coef_nb_bits = 13;
	else if (total_gain < 32)
		coef_nb_bits = 12;
	else if (total_gain < 40)
		coef_nb_bits = 11;
	else if (total_gain < 45)
		coef_nb_bits = 10;
	else
		coef_nb_bits = 9;

	/* compute number of coefficients */
	n = s->coefs_end[bsize] - s->coefs_start;
	for(ch = 0; ch < s->nb_channels; ch++)
		nb_coefs[ch] = n;

	/* complex coding */
	if (s->use_noise_coding) {

		for(ch = 0; ch < s->nb_channels; ch++) {
			if (s->channel_coded[ch]) {
				int i, n, a;
				n = s->exponent_high_sizes[bsize];
				for(i=0;i<n;i++) {
										if((i & 0xf) == 0) bitload(&s->gb);
					a = get_bits(&s->gb, 1);
					s->high_band_coded[ch][i] = a;
					/* if noise coding, the coefficients are not transmitted */
					if (a)
						nb_coefs[ch] -= s->exponent_high_bands[bsize][i];
				}
			}
		}
		for(ch = 0; ch < s->nb_channels; ch++) {
			if (s->channel_coded[ch]) {
				int i, n, val, code;

				n = s->exponent_high_sizes[bsize];
				val = (int)0x80000000;
				for(i=0;i<n;i++) {
					if (s->high_band_coded[ch][i]) {
						if (val == (int)0x80000000) {
													bitload(&s->gb);
							val = get_bits(&s->gb, 7) - 19;
						} else {
							code = get_vlc(&s->gb, &s->hgain_vlc);
							if (code < 0)
								return -1;
							val += code - 18;
						}
						s->high_band_values[ch][i] = val;
					}
				}
			}
		}
	}

	/* exposant can be interpolated in short blocks. */
	parse_exponents = 1;
	if (s->block_len_bits != s->frame_len_bits) {
		bitload(&s->gb);
		parse_exponents = get_bits(&s->gb, 1);
	}

	if (parse_exponents) {
		for(ch = 0; ch < s->nb_channels; ch++) {
			if (s->channel_coded[ch]) {
				if (s->use_exp_vlc) {
					if (decode_exp_vlc(s, ch) < 0)
						return -1;
				} else {
					decode_exp_lsp(s, ch);
				}
			}
		}
	} else {
		for(ch = 0; ch < s->nb_channels; ch++) {
			if (s->channel_coded[ch]) {
				interpolate_array(s->exponents[ch], 1 << s->prev_block_len_bits,
								  s->block_len);
			}
		}
	}

	/* parse spectral coefficients : just RLE encoding */
	for(ch = 0; ch < s->nb_channels; ch++) {
		if (s->channel_coded[ch]) {
			VLC *coef_vlc;
			int level, run, sign, tindex;
			int16_t *ptr, *eptr;
			const int16_t *level_table, *run_table;

			/* special VLC tables are used for ms stereo because
			   there is potentially less energy there */
			tindex = (ch == 1 && s->ms_stereo);
			coef_vlc = &s->coef_vlc[tindex];
			run_table = (const int16_t *)s->run_table[tindex];
			level_table = (const int16_t *)s->level_table[tindex];
			/* XXX: optimize */
			ptr = &s->coefs1[ch][0];
			eptr = ptr + nb_coefs[ch];
			memset(ptr, 0, s->block_len * sizeof(int16_t));
			for(;;) {
				code = get_vlc(&s->gb, coef_vlc);
				if (code < 0)
					return -1;
				if (code == 1) {
					/* EOB */
					break;
				} else if (code == 0) {
					/* escape */
					level = get_bits(&s->gb, coef_nb_bits);
					/* NOTE: this is rather suboptimal. reading
					   block_len_bits would be better */
					run = get_bits(&s->gb, s->frame_len_bits);
				} else {
					/* normal code */
					run = run_table[code];
					level = level_table[code];
				}
				sign = get_bits(&s->gb, 1);
				if (!sign)
					level = -level;
				ptr += run;
				if (ptr >= eptr)
					return -1;
				*ptr++ = level;
				/* NOTE: EOB can be omitted */
				if (ptr >= eptr)
					break;
			}
		}
		if (s->version == 1 && s->nb_channels >= 2) {
			bitload(&s->gb);
			align_get_bits(&s->gb);
		}
	}

	/* normalize */
	{
		int n4 = s->block_len / 2;
		mdct_norm = n4;
		if (s->version == 1) {
			mdct_norm /= ff_sqrt(n4);
		}
	}

	/* finally compute the MDCT coefficients */
	for(ch = 0; ch < s->nb_channels; ch++) {
		if (s->channel_coded[ch]) {
			int16_t *coefs1;
			int32_t *coefs;
#ifdef USEFLOAT
			float *exponents, *exp_ptr;
			float mult;
#else
			uint32_t *exponents, *exp_ptr;
			uint32_t mult;
#endif
			int i, j, n, n1, last_high_band;

			coefs1 = s->coefs1[ch];
			exponents = s->exponents[ch];

#ifdef USEFLOAT
			mult = internal_05_power_ten(total_gain) * s->max_exponent[ch] / mdct_norm;
#else
			mult = IRINT(internal_05_power_ten(total_gain) * s->max_exponent[ch], 11) / mdct_norm;
#endif

			coefs = s->coefs[ch];
			if (s->use_noise_coding) {
#ifdef USEFLOAT
				float exp_power[HIGH_BAND_MAX_SIZE];
				float mult1;
#else
				uint32_t exp_power[HIGH_BAND_MAX_SIZE];
				uint32_t mult1;
#endif

				/* very low freqs : noise */
				for(i = 0;i < s->coefs_start; i++) {
#ifdef USEFLOAT
					*coefs++ = lrintf(s->noise_table[s->noise_index] * (*exponents++) * mult);
#else
					*coefs++ = IRINT(IRINT((int64_t)s->noise_table[s->noise_index] * *exponents, 23) * mult, 32);
					exponents++;
#endif
					CONDTRIG("very low freqs : noise");
					s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
				}

				n1 = s->exponent_high_sizes[bsize];

				/* compute power of high bands */
				exp_ptr = exponents +
					s->high_band_start[bsize] -
					s->coefs_start;
				last_high_band = 0; /* avoid warning */
				for(j=0;j<n1;j++) {
					n = s->exponent_high_bands[s->frame_len_bits -
											  s->block_len_bits][j];
					if (s->high_band_coded[ch][j]) {
#ifdef USEFLOAT
						float e2, v;
#else
						uint32_t e2, v;
#endif
						e2 = 0;
						for(i = 0;i < n; i++) {
#ifdef USEFLOAT
							v = exp_ptr[i];
#else
							v = exp_ptr[i] >> 13;
#endif
							e2 += v * v;
						}
						exp_power[j] = e2 / n;
						last_high_band = j;
#ifdef USEFLOAT
						dprintf("%d: power=%f (%d)\n", j, exp_power[j], n);
#endif
						CONDTRIG("high_band_coded");
					}
					exp_ptr += n;
				}

				/* main freqs and high freqs */
				for(j=-1;j<n1;j++) {
#ifdef USEFLOAT
					float noise;
#else
					int32_t noise;
#endif
					if (j < 0) {
						n = s->high_band_start[bsize] -
							s->coefs_start;
					} else {
						n = s->exponent_high_bands[s->frame_len_bits -
												  s->block_len_bits][j];
					}
					if (j >= 0 && s->high_band_coded[ch][j]) {
#ifdef USEFLOAT
						/* use noise with specified power */
						mult1 = ff_sqrt(exp_power[j] / exp_power[last_high_band]);
						/* replaced mult1 pow() calculation with pre calculated values */
						mult1 = mult1 * internal_05_power_ten(s->high_band_values[ch][j]);
						mult1 = mult1 * s->max_exponent[ch] * s->noise_mult / mdct_norm;
#else
						/* use noise with specified power */
						/* replaced mult1 pow() calculation with pre calculated values */
						mult1 = IRINT(internal_05_power_ten(s->high_band_values[ch][j]) * s->max_exponent[ch], 15);
						mult1 = (uint64_t)mult1 * s->noise_mult * ff_sqrt(exp_power[j] / exp_power[last_high_band]) / mdct_norm;
#endif
						for(i = 0;i < n; i++) {
							noise = s->noise_table[s->noise_index];
#ifdef USEFLOAT
							*coefs++ = lrintf((*exponents++) * noise * mult1);
#else
							*coefs++ = IRINT(IRINT((int64_t)noise * *exponents, 19) * mult1, 32);
							exponents++;
#endif
							CONDTRIG("noise with specified power");
							s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
						}
					} else {
						/* coded values + small noise */
						for(i = 0;i < n; i++) {
							noise = s->noise_table[s->noise_index];
#ifdef USEFLOAT
							*coefs++ = lrintf((((int32_t)(*coefs1++) << 9) + noise) * (*exponents++) * mult);
#else
							*coefs++ = IRINT((((int64_t)*coefs1++ * *exponents) + IRINT((int64_t)noise * *exponents, 25)) * mult, 30);
							exponents++;
#endif
							CONDTRIG("coded values + small noise");
							s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
						}
					}
				}

				/* very high freqs : noise */
				n = s->block_len - s->coefs_end[bsize];
#ifdef USEFLOAT
				mult1 = mult * exponents[-1];
#else
				mult1 = IRINT((int64_t)mult * exponents[-1], 23); /* 20+19-16 */
#endif
				for(i = 0; i < n; i++) {
#ifdef USEFLOAT
					*coefs++ = lrintf(s->noise_table[s->noise_index] * mult1);
#else
					*coefs++ = IRINT((int64_t)s->noise_table[s->noise_index] * mult1, 32);
#endif
					CONDTRIG("very high freqs : noise");
					s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
				}
			} else {
				/* XXX: optimize more */
				for(i = 0;i < s->coefs_start; i++)
					*coefs++ = 0;
				n = nb_coefs[ch];
				for(i = 0;i < n; i++) {
#ifdef USEFLOAT
					*coefs++ = lrintf(((int32_t)coefs1[i] << 9) * exponents[i] * mult);
#else
					*coefs++ = IRINT(((int64_t)coefs1[i] * exponents[i]) * mult, 30);
#endif
					CONDTRIG("no noise coding");
				}
				n = s->block_len - s->coefs_end[bsize];
				for(i = 0;i < n; i++)
					*coefs++ = 0;
			}
		}
	}

	if (s->ms_stereo && s->channel_coded[1]) {
		int32_t a, b;
		int i;

		/* nominal case for ms stereo: we do it before mdct */
		/* no need to optimize this case because it should almost
		   never happen */
		if (!s->channel_coded[0]) {
			dprintf("rare ms-stereo case happened\n");
			memset(s->coefs[0], 0, sizeof(int32_t) * s->block_len);
			s->channel_coded[0] = 1;
		}

		for(i = 0; i < s->block_len; i++) {
			a = s->coefs[0][i];
			b = s->coefs[1][i];
			s->coefs[0][i] = a + b;
			s->coefs[1][i] = a - b;
		}
	}

	/* build the window : we ensure that when the windows overlap
	   their squared sum is always 1 (MDCT reconstruction rule) */
	/* XXX: merge with output */
	{
		int i, next_block_len, block_len, prev_block_len, n;
		int32_t *wptr;

		block_len = s->block_len;
		prev_block_len = 1 << s->prev_block_len_bits;
		next_block_len = 1 << s->next_block_len_bits;

		/* right part */
		wptr = window + block_len;
		if (block_len <= next_block_len) {
			for(i=0;i<block_len;i++)
				*wptr++ = s->windows[bsize][i];
		} else {
			/* overlap */
			n = (block_len / 2) - (next_block_len / 2);
			for(i=0;i<n;i++)
				*wptr++ = -0x80000000;
			for(i=0;i<next_block_len;i++)
				*wptr++ = s->windows[s->frame_len_bits - s->next_block_len_bits][i];
			for(i=0;i<n;i++)
				*wptr++ = 0;
		}

		/* left part */
		wptr = window + block_len;
		if (block_len <= prev_block_len) {
			for(i=0;i<block_len;i++)
				*--wptr = s->windows[bsize][i];
		} else {
			/* overlap */
			n = (block_len / 2) - (prev_block_len / 2);
			for(i=0;i<n;i++)
				*--wptr = -0x80000000;
			for(i=0;i<prev_block_len;i++)
				*--wptr = s->windows[s->frame_len_bits - s->prev_block_len_bits][i];
			for(i=0;i<n;i++)
				*--wptr = 0;
		}
	}


	for(ch = 0; ch < s->nb_channels; ch++) {
		if (s->channel_coded[ch]) {
			int32_t output[BLOCK_MAX_SIZE * 2] __attribute__((aligned(16)));
			int32_t *ptr;
			int i, n4, index, n;

			n = s->block_len;
			n4 = s->block_len >> 1;
			//printf("bl: %d\n", n);

			mdct_backward(n*2, s->coefs[ch], output);

			/* XXX: optimize all that by build the window and
			   multipying/adding at the same time */
			/* multiply by the window */
			for(i=0;i<n * 2;i++) {
				output[i] = FXSMULT(output[i], window[i], 31);
			}

			/* add in the frame */
			index = (s->frame_len / 2) + s->block_pos - n4;
			ptr = &s->frame_out[ch][index];
			for(i=0;i<n * 2;i++) {
				*ptr += output[i];
				ptr++;
			}

			/* specific fast case for ms-stereo : add to second
			   channel if it is not coded */
			if (s->ms_stereo && !s->channel_coded[1]) {
				ptr = &s->frame_out[1][index];
				for(i=0;i<n * 2;i++) {
					*ptr += output[i];
					ptr++;
				}
			}
		}
	}
 next:
	/* update block number */
	s->block_num++;
	s->block_pos += s->block_len;
	if (s->block_pos >= s->frame_len)
		return 1;
	else
		return 0;
}

/* decode a frame of frame_len samples */
static int wma_decode_frame(WMADecodeContext *s, int16_t *samples)
{
	int ret, i, n, a, ch, incr;
	int16_t *ptr;
	int32_t *iptr;

	/* read each block */
	s->block_num = 0;
	s->block_pos = 0;
	for(;;) {
		ret = wma_decode_block(s);
		if (ret < 0) {
			return -1;
		}
		if (ret)
			break;
	}


	/* convert frame to integer */
	n = s->frame_len;
	incr = s->nb_channels;
	for(ch = 0; ch < s->nb_channels; ch++) {

		ptr = samples + ch;
		iptr = s->frame_out[ch];

		for(i=0;i<n;i++) {

			a = IRINT(*iptr++, 9);

			if (a > 32767)
				a = 32767;
			else if (a < -32768)
				a = -32768;
			*ptr = a;

			ptr += incr;

		}

		/* prepare for next block */
		memmove(&s->frame_out[ch][0], &s->frame_out[ch][s->frame_len],
				s->frame_len * sizeof(int32_t));
		/* XXX: suppress this */
		memset(&s->frame_out[ch][s->frame_len], 0,
			   s->frame_len * sizeof(int32_t));
	}

	return 0;
}

int wma_decode_superframe(void *ctx,
								 void *data, int *data_size,
								 uint8_t *buf, int buf_size)
{
		WMADecodeContext *s = (WMADecodeContext *)ctx;
	int nb_frames, bit_offset, i, pos, len;
	uint8_t *q;
	int16_t *samples;

	if(s == NULL) return -1;

	dprintf("***decode_superframe:\n");

	if(buf_size==0){
		s->last_superframe_len = 0;
		return 0;
	}

	samples = (int16_t *)data;

	init_get_bits(&s->gb, buf, buf_size*8);

	if (s->use_bit_reservoir) {
		/* read super frame header */
		skip_bits(&s->gb, 4); /* super frame index */
		nb_frames = get_bits(&s->gb, 4) - 1;
			bitload(&s->gb);
		bit_offset = get_bits(&s->gb, s->byte_offset_bits + 3);

		if (s->last_superframe_len > 0) {
			//        printf("skip=%d\n", s->last_bitoffset);
			/* add bit_offset bits to last frame */
			if ((s->last_superframe_len + ((bit_offset + 7) >> 3)) >
				MAX_CODED_SUPERFRAME_SIZE)
				goto fail;
			q = s->last_superframe + s->last_superframe_len;
			len = bit_offset;
			while (len > 0) {
							bitload(&s->gb);
				*q++ = get_bits(&s->gb, 8);
				len -= 8;
			}
			if (len > 0) {
							bitload(&s->gb);
				*q++ = get_bits(&s->gb, len) << (8 - len);
			}

			/* XXX: bit_offset bits into last frame */
			init_get_bits(&s->gb, s->last_superframe, MAX_CODED_SUPERFRAME_SIZE*8);
			/* skip unused bits */
			if (s->last_bitoffset > 0)
				skip_bits(&s->gb, s->last_bitoffset);
			/* this frame is stored in the last superframe and in the
			   current one */
			if (wma_decode_frame(s, samples) < 0)
				goto fail;
			samples += s->nb_channels * s->frame_len;
		}

		/* read each frame starting from bit_offset */
		pos = bit_offset + 4 + 4 + s->byte_offset_bits + 3;
		init_get_bits(&s->gb, buf + (pos >> 3), (MAX_CODED_SUPERFRAME_SIZE - (pos >> 3))*8);
		len = pos & 7;
		if (len > 0)
			skip_bits(&s->gb, len);

		s->reset_block_lengths = 1;
		for(i=0;i<nb_frames;i++) {
			if (wma_decode_frame(s, samples) < 0)
				goto fail;
			samples += s->nb_channels * s->frame_len;
		}

		/* we copy the end of the frame in the last frame buffer */
#ifdef TCPMP
		bitload(&s->gb);
		s->last_bitoffset = s->gb.bitpos;
		pos = bitbytepos(&s->gb) - buf;
		if(s->last_bitoffset) pos--;
#else
		pos = get_bits_count(&s->gb) + ((bit_offset + 4 + 4 + s->byte_offset_bits + 3) & ~7);
		s->last_bitoffset = pos & 7;
		pos >>= 3;
#endif
		len = buf_size - pos;
		if (len > MAX_CODED_SUPERFRAME_SIZE || len < 0) {
			goto fail;
		}
		s->last_superframe_len = len;
		memcpy(s->last_superframe, buf + pos, len);
	} else {
		/* single frame decode */
		int count = buf_size / s->block_align;
		while(count--) {
			init_get_bits(&s->gb, buf, s->block_align*8);
			if (wma_decode_frame(s, samples) < 0)
				goto fail;
			samples += s->nb_channels * s->frame_len;
			buf += s->block_align;
		}
	}
	*data_size = (int8_t *)samples - (int8_t *)data;
	return s->block_align;
 fail:
	/* when error, we reset the bit reservoir */
	s->last_superframe_len = 0;
	return -1;
}

int wma_decode_end(void *ctx)
{
	WMADecodeContext *s = (WMADecodeContext *)ctx;
	int i;

	for(i = 0; i < s->nb_block_sizes; i++)
		free(s->windows[i]);

	if (s->use_exp_vlc) {
		free_vlc(&s->exp_vlc);
	}
	if (s->use_noise_coding) {
		free_vlc(&s->hgain_vlc);
	}
	for(i = 0;i < 2; i++) {
		free_vlc(&s->coef_vlc[i]);
		free(s->run_table[i]);
		free(s->level_table[i]);
	}

	free(s);

	return 0;
}
